(*
 * SNU 4190.310 Programming Languages 2017 Fall
 * Type Checker Skeleton
 *)

open M

type var = string

type typ =
  | TInt
  | TBool
  | TString
  | TPair of typ * typ
  | TLoc of typ
  | TFun of typ * typ
  | TVar of var
  (* Modify, or add more if needed *)

type typ_scheme =
  | SimpleTyp0 of typ
  | SimpleTyp1 of typ
  | GenTyp0 of (var list * typ)
  | GenTyp1 of (var list * typ)

type typ_env = (M.id * typ_scheme) list

let count = ref 0

let new_var () =
  let _ = count := !count + 1 in
  "x_" ^ (string_of_int !count)

(* My own definitions *)
let var_restrictions : (var, int) Hashtbl.t = Hashtbl.create 20 (* 0 : none, 1 : comparable, 2 : writable *)

let make_restriction : var -> int -> unit = fun x v ->
try (
  let now_val = Hashtbl.find var_restrictions x in
  if now_val < v then Hashtbl.replace var_restrictions x v else ()
) with Not_found -> Hashtbl.add var_restrictions x v

let find_restriction : var -> int = fun x ->
try (
  Hashtbl.find var_restrictions x
) with Not_found -> 0

let rec is_comparable : typ -> bool = fun ty -> match ty with
| TInt | TBool | TString -> true
| TVar v -> let _ = make_restriction v 1 in true
| TLoc ty' -> is_comparable ty'
| _ -> false

let rec is_writable : typ -> bool = fun ty -> match ty with
| TInt | TBool | TString -> true
| TVar v -> let _ = make_restriction v 2 in true
| _ -> false

(* Definitions related to free type variable *)

let union_ftv ftv_1 ftv_2 =
  let ftv_1' = List.filter (fun v -> not (List.mem v ftv_2)) ftv_1 in
  ftv_1' @ ftv_2

let sub_ftv ftv_1 ftv_2 =
  List.filter (fun v -> not (List.mem v ftv_2)) ftv_1

let rec ftv_of_typ : typ -> var list = function
  | TInt | TBool | TString -> []
  | TPair (t1, t2) -> union_ftv (ftv_of_typ t1) (ftv_of_typ t2)
  | TLoc t -> ftv_of_typ t
  | TFun (t1, t2) ->  union_ftv (ftv_of_typ t1) (ftv_of_typ t2)
  | TVar v -> [v]

let ftv_of_scheme : typ_scheme -> var list = function
  | SimpleTyp0 t -> ftv_of_typ t
  | SimpleTyp1 t -> ftv_of_typ t
  | GenTyp0 (alphas, t) -> sub_ftv (ftv_of_typ t) alphas
  | GenTyp1 (alphas, t) -> sub_ftv (ftv_of_typ t) alphas

let ftv_of_env : typ_env -> var list = fun tyenv ->
  List.fold_left
    (fun acc_ftv (id, tyscm) -> union_ftv acc_ftv (ftv_of_scheme tyscm))
    [] tyenv

(* Generalize given type into a type scheme *)
let generalize : typ_env -> typ -> typ_scheme = fun tyenv t ->
  let env_ftv = ftv_of_env tyenv in
  let typ_ftv = ftv_of_typ t in
  let ftv = sub_ftv typ_ftv env_ftv in
  if List.length ftv = 0 then
    SimpleTyp0 t
  else
    GenTyp0 (ftv, t)

(* Definitions related to substitution *)

type subst = typ -> typ

let empty_subst : subst = fun t -> t

let make_subst : var -> typ -> subst = fun x t ->
  let rec subs t' =
    match t' with
    | TVar x' -> if (x = x') then t else t'
    | TPair (t1, t2) -> TPair (subs t1, subs t2)
    | TLoc t'' -> TLoc (subs t'')
    | TFun (t1, t2) -> TFun (subs t1, subs t2)
    | TInt | TBool | TString -> t'
  in subs

let (@@) s1 s2 = (fun t -> s1 (s2 t)) (* substitution composition *)

let subst_scheme : subst -> typ_scheme -> typ_scheme = fun subs tyscm ->
  match tyscm with
  | SimpleTyp0 t -> SimpleTyp0 (subs t)
  | SimpleTyp1 t -> SimpleTyp1 (subs t)
  | GenTyp0 (alphas, t) | GenTyp1 (alphas, t) ->
    (* S (\all a.t) = \all b.S{a->b}t  (where b is new variable) *)
    let betas = List.map (fun orig_name ->
      let new_name = new_var () in
      let _ = make_restriction new_name (find_restriction orig_name) in
      new_name
    ) alphas in
    let s' =
      List.fold_left2
        (fun acc_subst alpha beta -> make_subst alpha (TVar beta) @@ acc_subst)
        empty_subst alphas betas
    in
    match tyscm with
    | GenTyp0 _ -> GenTyp0 (betas, subs (s' t))
    | GenTyp1 _ -> GenTyp1 (betas, subs (s' t))
    | _ -> failwith "what?"

let subst_env : subst -> typ_env -> typ_env = fun subs tyenv ->
  List.map (fun (x, tyscm) -> (x, subst_scheme subs tyscm)) tyenv

(* debugging function *)
let rec type_string : typ -> string = fun x -> match x with
| TInt -> "int"
| TBool -> "bool"
| TString -> "string"
| TPair (u, v) -> "(" ^ (type_string u) ^ ", " ^ (type_string v) ^ ")"
| TLoc u -> (type_string u) ^ "*"
| TFun (u, v) -> "(" ^ (type_string u) ^ " -> " ^ (type_string v) ^ ")"
| TVar n ->
  let rlevel = find_restriction n in
  if (rlevel = 0) then n
  else if (rlevel = 1) then ("cpext " ^ n)
  else if (rlevel = 2) then ("wrext " ^ n)
  else raise (M.TypeError "how could this even happen?")

(* TODO : Implement this function *)
let check : M.exp -> M.typ = fun exp ->
  let rec is_in : typ -> var -> bool = fun tau alpha -> match tau with
  | TInt | TBool | TString -> false
  | TPair (u, v) -> if (is_in u alpha) then true else (is_in v alpha)
  | TLoc u -> is_in u alpha
  | TFun (u, v) -> if (is_in u alpha) then true else (is_in v alpha)
  | TVar u -> u = alpha
  in
  (* CORRECT? *)
  let rec unify : typ -> typ -> subst = fun tau1 tau2 -> match (tau1, tau2) with
  | (TVar v, tau) | (tau, TVar v) ->
    if (is_in tau v) && not (tau1 = tau2) then (
      raise (M.TypeError ("cannot unify " ^ (type_string tau1) ^ " and " ^ (type_string tau2)))
    ) else (
      let rlevel = (
      match tau with
      | TVar u ->
        let l1 = find_restriction v in
        let l2 = find_restriction u in
        if l1 > l2 then l1 else l2
      | _ -> find_restriction v
      ) in
      if (rlevel = 0) then (
        make_subst v tau
      ) else if (rlevel = 1) then (
        if (is_comparable tau) then (
          make_subst v tau
        ) else raise (M.TypeError (
          "cannot unify " ^ (type_string tau1) ^ " and " ^ (type_string tau2)
        ))
      ) else if (rlevel = 2) then (
        if (is_writable tau) then (
          make_subst v tau
        ) else raise (M.TypeError (
          "cannot unify " ^ (type_string tau1) ^ " and " ^ (type_string tau2)
        ))
      ) else raise (M.TypeError "how could this even happen?")
    )
  | (TPair (t1, t2), TPair (t1', t2')) ->
    let s = unify t1 t1' in
    let s' = unify (s t2) (s t2') in
    s' @@ s
  | (TLoc u, TLoc v) -> unify u v
  | (TFun (t1, t2), TFun (t1', t2')) ->
    let s = unify t1 t1' in
    let s' = unify (s t2) (s t2') in
    s' @@ s
  | _ ->
    if (tau1 = tau2) then empty_subst else raise (M.TypeError (
      "cannot unify " ^ (type_string tau1) ^ " and " ^ (type_string tau2)
    ))
  in
  (* I have zero confidence... *)
  let rec expansive : typ_env -> M.exp -> int = fun tenv exp ->
    let rec expansive_inner ls exp = match exp with
    | M.CONST _ -> 0
    | M.VAR v ->
      let exp_level = ls v in
      if (exp_level >= 0) then exp_level else (
        try (
          let u = snd (List.find (fun (x, _) -> v = x) tenv) in
          let in_lvl = match u with
          | SimpleTyp0 _ -> 0
          | SimpleTyp1 _ -> 1
          | GenTyp0 _ -> 0
          | GenTyp1 _ -> 1
          in
          let _ = print_endline ("expansive: " ^ v ^ " = " ^ (string_of_int in_lvl)) in
          in_lvl
        ) with Not_found -> 2
      )
    | M.FN (v, e) ->
      let exp_level = expansive_inner (fun x -> if (v = x) then 0 else (ls x)) e in
      if (exp_level = 2) then 1 else exp_level
    | M.APP (e1, e2) ->
      if (expansive_inner ls e1) = 0 && (expansive_inner ls e2) = 0 then 0 else 2
    | M.LET (M.VAL (_, exp1), exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.LET (M.REC (_, _, exp1), exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.IF (cond, exp1, exp2) -> if (expansive_inner ls cond) = 0 && (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.BOP (_, exp1, exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.READ -> 0
    | M.WRITE exp' -> expansive_inner ls exp'
    | M.MALLOC _ -> 2
    | M.ASSIGN (exp1, exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.BANG exp' -> expansive_inner ls exp'
    | M.SEQ (exp1, exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.PAIR (exp1, exp2) -> if (expansive_inner ls exp1) = 0 && (expansive_inner ls exp2) = 0 then 0 else 2
    | M.FST exp' -> expansive_inner ls exp'
    | M.SND exp' -> expansive_inner ls exp'
    in expansive_inner (fun x -> -1) exp
  in
  (* It is working on all of my test cases but *)
  let rec check_env : typ_env -> M.exp -> typ -> subst = fun tenv exp tau -> match exp with
  | M.CONST c -> (
    match c with
    | M.N _ -> unify tau TInt
    | M.B _ -> unify tau TBool
    | M.S _ -> unify tau TString
    )
  | M.VAR name ->
    let tau' = snd (List.find (fun x -> (fst x) = name) tenv) in
    let tau'' = subst_scheme empty_subst tau' in (
    match tau'' with
    | SimpleTyp0 tau''' -> unify tau tau'''
    | SimpleTyp1 tau''' -> unify tau tau'''
    | GenTyp0 (_, tau''') -> unify tau tau''' (* O RLY? *)
    | GenTyp1 (_, tau''') -> unify tau tau''' (* O RLY? *)
    )
  | M.FN (pvar, exp') ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let s1 = unify tau (TFun (beta1, beta2)) in
    let s2 = check_env ((pvar, SimpleTyp0 (s1 beta1))::(subst_env s1 tenv)) exp' (s1 beta2) in (* O RLY? *)
    s2 @@ s1
  | M.APP (exp1, exp2) ->
    let beta = TVar (new_var ()) in
    let s1 = check_env tenv exp1 (TFun (beta, tau)) in
    let s2 = check_env (subst_env s1 tenv) exp2 (s1 beta) in
    s2 @@ s1
  | M.LET (M.VAL (name, exp1), exp2) ->
    let beta = TVar (new_var ()) in
    let s1 = check_env tenv exp1 beta in
    let ntenv = subst_env s1 tenv in
    let exp_lvl = expansive ntenv exp1 in
    let _ = print_endline (name ^ " = " ^ (string_of_int exp_lvl)) in
    let scheme = if exp_lvl = 2 then (SimpleTyp1 (s1 beta)) else (
      let u = generalize ntenv (s1 beta) in
      if exp_lvl = 1 then (
        match u with
        | GenTyp0 (a, b) -> GenTyp1 (a, b)
        | SimpleTyp0 a -> SimpleTyp1 a
        | _ -> u
      ) else u
    ) in
    let s2 = check_env ((name, scheme) :: ntenv) exp2 (s1 tau) in
    s2 @@ s1
  | M.LET (M.REC (fname, name, exp1), exp2) ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let func_type = TFun (beta1, beta2) in
    let s1 = check_env ((name, SimpleTyp0 beta1) :: ((fname, SimpleTyp0 func_type) :: tenv)) exp1 beta2 in
    let ntenv = subst_env s1 tenv in
    let exp_lvl = expansive ((name, SimpleTyp0 (s1 beta1)) :: ((fname, SimpleTyp0 (s1 func_type)) :: ntenv)) exp1 in
    let _ = print_endline (fname ^ " = " ^ (string_of_int exp_lvl)) in
    let scheme = (
      let u = generalize ntenv (s1 func_type) in
      if exp_lvl >= 1 then (
        match u with
        | GenTyp0 (a, b) -> GenTyp1 (a, b)
        | SimpleTyp0 a -> SimpleTyp1 a
        | _ -> u
      ) else u
    ) in
    let s2 = check_env ((fname, scheme) :: ntenv) exp2 (s1 tau) in
    s2 @@ s1
  | M.IF (cond, exp1, exp2) ->
    let beta = TVar (new_var ()) in
    let s1 = unify tau beta in
    let s2 = check_env (subst_env s1 tenv) cond TBool in
    let s' = s2 @@ s1 in
    let s3 = check_env (subst_env s' tenv) exp1 (s' beta) in
    let s4 = check_env (subst_env (s3 @@ s') tenv) exp2 (s3 (s' beta)) in
    s4 @@ s3 @@ s'
  | M.BOP (bop, exp1, exp2) -> (
    match bop with
    | M.ADD | M.SUB ->
      let s1 = unify tau TInt in
      let s2 = check_env (subst_env s1 tenv) exp1 TInt in
      let s3 = check_env (subst_env (s1 @@ s2) tenv) exp2 TInt in
      s3 @@ s2 @@ s1
    | M.EQ ->
      let beta = TVar (new_var ()) in
      let s1 = unify tau TBool in
      let s2 = check_env (subst_env s1 tenv) exp1 (s1 beta) in
      let s' = s2 @@ s1 in
      let s3 = check_env (subst_env s' tenv) exp2 (s' beta) in
      let beta' = s3 (s' beta) in
      if (is_comparable beta') then (
        s3 @@ s'
      ) else (
        raise (M.TypeError ("cannot compare two variables with " ^ (type_string beta') ^ " type"))
      )
    | M.AND | M.OR ->
      let s1 = unify tau TBool in
      let s2 = check_env (subst_env s1 tenv) exp1 TBool in
      let s3 = check_env (subst_env (s1 @@ s2) tenv) exp2 TBool in
      s3 @@ s2 @@ s1
    )
  | M.READ -> unify tau TInt
  | M.WRITE exp' ->
    let beta = TVar (new_var ()) in
    let s1 = unify tau beta in
    let s2 = check_env (subst_env s1 tenv) exp' (s1 beta) in
    let beta' = s2 (s1 beta) in
    if (is_writable beta') then (
      s2 @@ s1
    ) else (
      raise (M.TypeError ("cannot print " ^ (type_string beta') ^ " type"))
    )
  | M.MALLOC exp' ->
    let beta = TVar (new_var ()) in
    let s1 = unify tau (TLoc beta) in
    let s2 = check_env (subst_env s1 tenv) exp' (s1 beta) in
    s2 @@ s1
  | M.ASSIGN (exp1, exp2) ->
    let s1 = check_env tenv exp1 (TLoc tau) in
    let s2 = check_env (subst_env s1 tenv) exp2 (s1 tau) in
    s2 @@ s1
  | M.BANG exp' -> check_env tenv exp' (TLoc tau)
  | M.SEQ (exp1, exp2) ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let s1 = unify tau beta2 in
    let s2 = check_env (subst_env s1 tenv) exp1 (s1 beta1) in
    let s3 = check_env (subst_env (s2 @@ s1) tenv) exp2 (s2 (s1 beta2)) in
    s3 @@ s2 @@ s1
  | M.PAIR (exp1, exp2) ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let s1 = unify tau (TPair (beta1, beta2)) in
    let s2 = check_env (subst_env s1 tenv) exp1 (s1 beta1) in
    let s3 = check_env (subst_env (s2 @@ s1) tenv) exp2 (s2 (s1 beta2)) in
    s3 @@ s2 @@ s1
  | M.FST exp' ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let s1 = unify tau beta1 in
    let s2 = check_env (subst_env s1 tenv) exp' (TPair (s1 beta1, s1 beta2)) in
    s2 @@ s1
  | M.SND exp' ->
    let (beta1, beta2) = (TVar (new_var ()), TVar (new_var ())) in
    let s1 = unify tau beta2 in
    let s2 = check_env (subst_env s1 tenv) exp' (TPair (s1 beta1, s1 beta2)) in
    s2 @@ s1
  in
  let new_name = new_var () in
  let subs : subst = check_env [] exp (TVar new_name) in
  let rec convert_type : typ -> M.typ = fun t -> match t with
  | TInt -> M.TyInt
  | TBool -> M.TyBool
  | TString -> M.TyString
  | TPair (t1, t2) -> M.TyPair (convert_type t1, convert_type t2)
  | TLoc t' -> M.TyLoc (convert_type t')
  | _ -> raise (M.TypeError "cannot convert typ to M.typ")
  in
(*let _ = print_endline ("type_string: " ^  (type_string (subs (TVar new_name)))) in *)
  convert_type (subs (TVar new_name))
